from django.conf import settings
from django.utils.http import url_has_allowed_host_and_scheme
from django.shortcuts import redirect
from djangoldp.models import Model


class AllowOnlySiteUrl:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        if(url_has_allowed_host_and_scheme(request.get_raw_uri(), allowed_hosts=settings.SITE_URL) or response.status_code != 200):
            return response
        else:
            return redirect('{}{}'.format(settings.SITE_URL, request.path), permanent=True)


class AllowRequestedCORSMiddleware:
    '''A CORS Middleware which allows the domains requested by the request'''
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)

        response["Access-Control-Allow-Origin"] = request.headers.get('origin')
        response["Access-Control-Allow-Methods"] = "GET,POST,PUT,PATCH,DELETE,OPTIONS,HEAD"
        response["Access-Control-Allow-Headers"] = \
            getattr(settings, 'OIDC_ACCESS_CONTROL_ALLOW_HEADERS',
                    "authorization, Content-Type, if-match, accept, DPoP, cache-control, prefer")
        response["Access-Control-Expose-Headers"] = "Location, User"
        response["Access-Control-Allow-Credentials"] = 'true'

        return response

class PrefetchParentModel:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        try:
            parent_model = Model.resolve_parent(request.path)
        except:
            parent_model = None
        request.parent_model = parent_model
        return self.get_response(request)


"""This module override some of default django configuration from the global settings."""

from django.conf.global_settings import *

####################
# LDP              #
####################

LDP_RDF_CONTEXT = 'https://cdn.happy-dev.fr/owl/hdcontext.jsonld'

MAX_ACTIVITY_RESCHEDULES = 3
DEFAULT_BACKOFF_FACTOR = 1
DEFAULT_ACTIVITY_DELAY = 3
DEFAULT_REQUEST_TIMEOUT = 10

####################
# CORE             #
####################

# https://en.wikipedia.org/wiki/List_of_tz_zones_by_name (although not all
# systems may support all possibilities). When USE_TZ is True, this is
# interpreted as the default user time zone.
TIME_ZONE = 'UTC'

# If you set this to True, Django will use timezone-aware datetimes.
USE_TZ = True

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

# Database connection info. If left empty, will default to the dummy backend.
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db.sqlite3',
    }
}

# List of strings representing installed apps.
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'djangoldp',
    'guardian'
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# Default server url
SITE_URL = 'http://localhost:8000'
BASE_URL = SITE_URL

# Default URL conf
ROOT_URLCONF = 'djangoldp.urls'

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/var/www/example.com/media/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT.
# Examples: "http://example.com/media/", "http://media.example.com/"
MEDIA_URL = '/media/'

# Absolute path to the directory static files should be collected to.
# Example: "/var/www/example.com/static/"
STATIC_ROOT = None

# URL that handles the static files served from STATIC_ROOT.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/static/'

# ETAGS config
USE_ETAGS = True

# Default X-Frame-Options header value
X_FRAME_OPTIONS = 'DENY'

# The Python dotted path to the WSGI application that Django's internal server
# (runserver) will use. If `None`, the return value of
# 'django.core.wsgi.get_wsgi_application' is used, thus preserving the same
# behavior as previous versions of Django. Otherwise this should point to an
# actual WSGI application object.
WSGI_APPLICATION = 'server.wsgi.application'

##############
# MIDDLEWARE #
##############

# List of middleware to use. Order is important; in the request phase, these
# middleware will be applied in the order given, and in the response
# phase the middleware will be applied in reverse order.
MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'djangoldp.middleware.AllowRequestedCORSMiddleware',
    'djangoldp.middleware.PrefetchParentModel',
    'django.middleware.gzip.GZipMiddleware',
    'django_brotli.middleware.BrotliMiddleware'
]

##################
# REST FRAMEWORK #
##################
REST_FRAMEWORK = {
    'DEFAULT_SCHEMA_CLASS': 'drf_spectacular.openapi.AutoSchema',
    'DEFAULT_PAGINATION_CLASS': 'djangoldp.pagination.LDPPagination'
}

###################
# DRF SPECTACULAR #
###################
ENABLE_SWAGGER_DOCUMENTATION = False
SPECTACULAR_SETTINGS = {
    'TITLE': 'DjangoLDP Based API Description',
    'DESCRIPTION': 'Here you will find the list of all endpoints available on your Djangoldp-based server instance and\
         the available methods, needed parameters and requests examples. The list of available endpoints depend on your instance configuration\
             especially the list of bloxes and associated django models activated.',
    'VERSION': '2.1.37',
    'SERVE_INCLUDE_SCHEMA': False
}

############
# SESSIONS #
############

# Same site policy
DCS_SESSION_COOKIE_SAMESITE = 'none'

##################
# AUTHENTICATION #
##################

AUTHENTICATION_BACKENDS = ['django.contrib.auth.backends.ModelBackend', 'guardian.backends.ObjectPermissionBackend']

OIDC_ACCESS_CONTROL_ALLOW_HEADERS = 'Content-Type, if-match, accept, authorization, DPoP, cache-control, pragma, prefer'

# The minimum number of seconds a password reset link is valid for
PASSWORD_RESET_TIMEOUT = 60 * 60 * 24 * 3

DISABLE_LOCAL_OBJECT_FILTER = False

GUARDIAN_AUTO_PREFETCH = True
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'


import uuid
from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import TestCase
from rest_framework.test import APIRequestFactory, APIClient
from rest_framework.utils import json

from djangoldp.tests.models import UserProfile, Resource, Invoice, Batch, Task, Skill, JobOffer, Conversation, Project,\
    NotificationSetting


class Update(TestCase):

    def setUp(self):
        self.factory = APIRequestFactory()
        self.client = APIClient()
        self.user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                         password='glass onion')
        self.client.force_authenticate(user=self.user)

    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/326
    '''
    def test_update_container_append_resource(self):
        pre_existing_skill_a = Skill.objects.create(title="to keep", obligatoire="obligatoire", slug="slug1")
        pre_existing_skill_b = Skill.objects.create(title="to keep", obligatoire="obligatoire", slug="slug2")
        job = JobOffer.objects.create(title="job test")
        job.skills.add(pre_existing_skill_a)
        job.skills.add(pre_existing_skill_b)

        post = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
                "skills": {
                    "ldp:contains": [
                        {"title": "new skill", "obligatoire": "okay"},
                        {"@id": "{}/skills/{}/".format(settings.BASE_URL, pre_existing_skill_b.slug), "title": "z"},
                    ]}
                }

        response = self.client.patch('/job-offers/{}/'.format(job.slug),
                                     data=json.dumps(post),
                                     content_type='application/ld+json')
        self.assertEquals(response.status_code, 200)

        self.assertEquals(response.data['title'], job.title)
        self.assertIs(job.skills.count(), 3)
        skills = job.skills.all().order_by('title')
        self.assertEquals(skills[0].title, "new skill")  # new skill
        self.assertEquals(skills[1].title, pre_existing_skill_a.title)  # old skill unchanged
        self.assertEquals(skills[2].title, "z")  # updated
        self.assertEquals(skills[2].obligatoire, pre_existing_skill_b.obligatoire)  # another field not updated
    '''

    def test_put_resource(self):
        skill = Skill.objects.create(title='original', obligatoire='original', slug='skill1')
        body = [{
            '@id': '{}/skills/{}/'.format(settings.BASE_URL, skill.slug),
            'http://happy-dev.fr/owl/#title': "new", 'http://happy-dev.fr/owl/#obligatoire': "new"}]
        response = self.client.put('/skills/{}/'.format(skill.slug), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(response.data['title'], "new")
        self.assertEquals(response.data['obligatoire'], "new")
        self.assertIn('location', response.headers)

    def test_patch_resource(self):
        skill = Skill.objects.create(title='original', obligatoire='original', slug='skill1')
        body = {
            '@id': '{}/skills/{}'.format(settings.BASE_URL, skill.slug),
            'http://happy-dev.fr/owl/#title': 'new'
        }
        response = self.client.patch('/skills/{}/'.format(skill.slug), data=json.dumps(body),
                                     content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(response.data['title'], "new")
        self.assertEquals(response.data['obligatoire'], "original")

    def test_create_sub_object_in_existing_object_with_existing_reverse_1to1_relation(self):
        user = get_user_model().objects.create(username="alex", password="test")
        profile = UserProfile.objects.create(user=user, description="user description")
        body = [
            {
                '@id': "/userprofiles/{}/".format(profile.pk),
                'http://happy-dev.fr/owl/#description': "user update"
            },
            {
                '@id': '/users/{}/'.format(user.pk),
                "http://happy-dev.fr/owl/#first_name": "Alexandre",
                "http://happy-dev.fr/owl/#last_name": "Bourlier",
                "http://happy-dev.fr/owl/#username": "alex",
                'http://happy-dev.fr/owl/#userprofile': {'@id': "/userprofiles/{}/".format(profile.pk)}
            }
        ]
        response = self.client.put('/users/{}/'.format(user.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('userprofile', response.data)

    def test_put_nonexistent_local_resource(self):
        job = JobOffer.objects.create(title="job test")

        # contains internal urlid which refers to non-existent resource
        body = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
                "skills": {
                    "ldp:contains": [
                        {"@id": "{}/skills/404/".format(settings.BASE_URL)},
                    ]}
                }

        response = self.client.put('/job-offers/{}/'.format(job.slug), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(Skill.objects.count(), 0)

    def test_create_sub_object_in_existing_object_with_reverse_fk_relation(self):
        """
        Doesn't work with depth = 0 on UserProfile Model. Should it be ?
        """
        user = get_user_model().objects.create(username="alex", password="test")
        body = [
            {
                '@id': "_:b975",
                'http://happy-dev.fr/owl/#description': "conversation description"
            },
            {
                '@id': '/users/{}/'.format(user.pk),
                "http://happy-dev.fr/owl/#first_name": "Alexandre",
                "http://happy-dev.fr/owl/#last_name": "Bourlier",
                "http://happy-dev.fr/owl/#username": "alex",
                'http://happy-dev.fr/owl/#conversation_set': {'@id': "_:b975"}
            }
        ]
        response = self.client.put('/users/{}/'.format(user.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('conversation_set', response.data)

    def test_create_sub_object_in_existing_object_with_existing_reverse_fk_relation(self):
        user = get_user_model().objects.create(username="alex", password="test")
        conversation = Conversation.objects.create(author_user=user, description="conversation description")
        body = [
            {
                '@id': "/conversations/{}/".format(conversation.pk),
                'http://happy-dev.fr/owl/#description': "conversation update"
            },
            {
                '@id': '/users/{}/'.format(user.pk),
                "http://happy-dev.fr/owl/#first_name": "Alexandre",
                "http://happy-dev.fr/owl/#last_name": "Bourlier",
                "http://happy-dev.fr/owl/#username": "alex",
                'http://happy-dev.fr/owl/#conversation_set': {'@id': "/conversations/{}/".format(conversation.pk)}
            }
        ]
        response = self.client.put('/users/{}/'.format(user.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('conversation_set', response.data)

    def test_missing_field_should_not_be_removed_with_fk_relation(self):
        peer = get_user_model().objects.create(username="sylvain", password="test2")
        conversation = Conversation.objects.create(author_user=self.user, peer_user=peer,
                                                   description="conversation description")
        body = [
            {
                '@id': "/conversations/{}/".format(conversation.pk),
                'http://happy-dev.fr/owl/#description': "conversation update",
            }
        ]
        response = self.client.put('/conversations/{}/'.format(conversation.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('peer_user', response.data)

    def test_empty_field_should_be_removed_with_fk_relation(self):
        peer = get_user_model().objects.create(username="sylvain", password="test2")
        conversation = Conversation.objects.create(author_user=self.user, peer_user=peer,
                                                   description="conversation description")
        body = [
            {
                '@id': "/conversations/{}/".format(conversation.pk),
                'http://happy-dev.fr/owl/#description': "conversation update",
                'http://happy-dev.fr/owl/#peer_user': ""
            }
        ]
        response = self.client.put('/conversations/{}/'.format(conversation.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['peer_user'], None)

    def test_m2m_new_link_bis(self):
        resource = Resource.objects.create()
        job = JobOffer.objects.create(title="first title", slug="job")
        body = {
            'http://happy-dev.fr/owl/#joboffers':
                {
                    '@id': "{}/resources/{}/joboffers/".format(settings.BASE_URL, resource.pk),
                    'ldp:contains': [
                        {'@id': job.urlid,
                         'http://happy-dev.fr/owl/#title': "new job",
                         },
                    ]
                }
        }

        response = self.client.put('/resources/{}/'.format(resource.pk),
                                   data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['joboffers']['ldp:contains'][0]['@id'], job.urlid)
        self.assertEqual(response.data['joboffers']['ldp:contains'][0]['title'], "new job")

    def test_m2m_new_link_embedded(self):
        resource = Resource.objects.create()
        body = {
            'http://happy-dev.fr/owl/#joboffers': {
                'http://happy-dev.fr/owl/#slug': 'aaa',
                'http://happy-dev.fr/owl/#title': "new job",
            }
        }

        response = self.client.put('/resources/{}/'.format(resource.pk),
                                   data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['joboffers']['ldp:contains'][0]['@id'],
                         "http://happy-dev.fr/job-offers/aaa/")
        self.assertEqual(response.data['joboffers']['ldp:contains'][0]['title'], "new job")

    def test_m2m_existing_link(self):
        resource = Resource.objects.create()
        job = JobOffer.objects.create(title="first title", slug="job")
        resource.joboffers.add(job)
        resource.save()
        body = {
            'http://happy-dev.fr/owl/#joboffers': {
                # '@id': "http://testserver/resources/{}/joboffers/".format(resource.pk),
                'ldp:contains': [
                    {
                        '@id': job.urlid,
                        'http://happy-dev.fr/owl/#title': "new job",
                    }
                ]
            }
        }

        response = self.client.put('/resources/{}/'.format(resource.pk),
                                   data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['joboffers']['ldp:contains'][0]['@id'], job.urlid)
        self.assertEqual(response.data['joboffers']['ldp:contains'][0]['title'], "new job")

    def test_m2m_new_link_external(self):
        resource = Resource.objects.create()
        body = {
            'http://happy-dev.fr/owl/#joboffers': {
                'http://happy-dev.fr/owl/#@id': 'http://external.job/job/1',
            }
        }

        response = self.client.put('/resources/{}/'.format(resource.pk),
                                   data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['joboffers']['ldp:contains'][0]['@id'],
                         "http://external.job/job/1")
        self.assertIn('@type', response.data['joboffers']['ldp:contains'][0])

    def test_m2m_new_link_local(self):
        resource = Resource.objects.create()
        job = JobOffer.objects.create(title="first title", slug="job")
        body = {
            'http://happy-dev.fr/owl/#joboffers': {
                '@id': 'http://happy-dev.fr/job-offers/{}/'.format(job.slug),
            }
        }

        response = self.client.put('/resources/{}/'.format(resource.pk),
                                   data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['joboffers']['ldp:contains'][0]['@id'],
                         "http://happy-dev.fr/job-offers/{}/".format(job.slug))
        self.assertEqual(response.data['joboffers']['ldp:contains'][0]['title'], "first title")

    def test_update_with_new_fk_relation(self):
        conversation = Conversation.objects.create(author_user=self.user,
                                                   description="conversation description")
        body = [
            {
                '@id': "/conversations/{}/".format(conversation.pk),
                'http://happy-dev.fr/owl/#description': "conversation update",
                'http://happy-dev.fr/owl/#peer_user': {
                    '@id': 'http://happy-dev.fr/users/{}'.format(self.user.pk),
                }
            }
        ]
        response = self.client.put('/conversations/{}/'.format(conversation.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('peer_user', response.data)

        conversation = Conversation.objects.get(pk=conversation.pk)
        self.assertIsNotNone(conversation.peer_user)

        user = get_user_model().objects.get(pk=self.user.pk)
        self.assertEqual(user.peers_conv.count(), 1)

    def test_m2m_user_link_federated(self):
        project = Project.objects.create(description="project name")
        body = {
            'http://happy-dev.fr/owl/#description': 'project name',
            'http://happy-dev.fr/owl/#members': {
                'http://happy-dev.fr/owl/#@id': 'http://external.user/user/1',
            }
        }

        response = self.client.put('/projects/{}/'.format(project.pk),
                                   data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['members']['ldp:contains'][0]['@id'],
                         "http://external.user/user/1")
        self.assertIn('@type', response.data['members']['ldp:contains'][0])
        self.assertEqual(len(response.data['members']['ldp:contains'][0].items()), 2)

    def test_m2m_user_link_existing_external(self):
        project = Project.objects.create(description="project name")
        ext_user = get_user_model().objects.create(username=str(uuid.uuid4()), urlid='http://external.user/user/1')
        body = {
            'http://happy-dev.fr/owl/#description': 'project name',
            'http://happy-dev.fr/owl/#members': {
                'http://happy-dev.fr/owl/#@id': ext_user.urlid,
            }
        }

        response = self.client.put('/projects/{}/'.format(project.pk),
                                   data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['members']['ldp:contains'][0]['@id'], ext_user.urlid)
        self.assertIn('@type', response.data['members']['ldp:contains'][0])
        self.assertEqual(len(response.data['members']['ldp:contains'][0].items()), 2)

        project = Project.objects.get(pk=project.pk)
        self.assertEqual(project.members.count(), 1)

        user = get_user_model().objects.get(pk=ext_user.pk)
        self.assertEqual(user.projects.count(), 1)

    def test_create_sub_object_in_existing_object_with_reverse_1to1_relation(self):
        """
        Doesn't work with depth = 0 on UserProfile Model. Should it be ?
        """
        user = get_user_model().objects.create(username="alex", password="test")
        body = [
            {
                '@id': "_:b975",
                'http://happy-dev.fr/owl/#description': "user description",
                'http://happy-dev.fr/owl/#dummy': {
                    '@id': './'
                }
            },
            {
                '@id': '/users/{}/'.format(user.pk),
                "http://happy-dev.fr/owl/#first_name": "Alexandre",
                "http://happy-dev.fr/owl/#last_name": "Bourlier",
                "http://happy-dev.fr/owl/#username": "alex",
                'http://happy-dev.fr/owl/#userprofile': {'@id': "_:b975"}
            }
        ]
        response = self.client.put('/users/{}/'.format(user.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('userprofile', response.data)
        self.assertIsNotNone(response.data['userprofile'])

    def test_m2m_user_link_remove_existing_link(self):
        ext_user = get_user_model().objects.create(username=str(uuid.uuid4()), urlid='http://external.user/user/1')
        project = Project.objects.create(description="project name")
        project.members.add(ext_user)
        project.save()
        body = {
            'http://happy-dev.fr/owl/#description': 'project name',
            'http://happy-dev.fr/owl/#members': {
            }
        }

        response = self.client.put('/projects/{}/'.format(project.pk),
                                   data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

        project = Project.objects.get(pk=project.pk)
        self.assertEqual(project.members.count(), 0)

        user = get_user_model().objects.get(pk=ext_user.pk)
        self.assertEqual(user.projects.count(), 0)

    def test_update_sub_object_with_urlid(self):
        user = get_user_model().objects.create(username="alex", password="test")
        profile = UserProfile.objects.create(user=user, description="user description")
        body = {
            '@id': '/users/{}/'.format(user.pk),
            "first_name": "Alexandre",
            "last_name": "Bourlier",
            "username": "alex",
            'userprofile': {
                '@id': profile.urlid,
                'description': "user update"
            },
            '@context': {
                "@vocab": "http://happy-dev.fr/owl/#",
            }
        }

        response = self.client.put('/users/{}/'.format(user.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('userprofile', response.data)

        response = self.client.get('/userprofiles/{}/'.format(profile.pk),
                                   content_type='application/ld+json')
        self.assertEqual(response.data['description'], "user update")

    # unit tests for a specific bug: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/307
    def test_direct_boolean_field(self):
        profile = UserProfile.objects.create(user=self.user)
        setting = NotificationSetting.objects.create(user=profile, receiveMail=False)
        body = {
            'http://happy-dev.fr/owl/#@id': setting.urlid,
            'receiveMail': True,
            "@context": {"@vocab": "http://happy-dev.fr/owl/#",
                         "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                         "rdfs": "http://www.w3.org/2000/01/rdf-schema#", "ldp": "http://www.w3.org/ns/ldp#",
                         "foaf": "http://xmlns.com/foaf/0.1/", "name": "rdfs:label",
                         "acl": "http://www.w3.org/ns/auth/acl#", "permissions": "acl:accessControl",
                         "mode": "acl:mode", "geo": "http://www.w3.org/2003/01/geo/wgs84_pos#", "lat": "geo:lat",
                         "lng": "geo:long"}
        }

        response = self.client.patch('/notificationsettings/{}/'.format(setting.pk),
                                     data=json.dumps(body),
                                     content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['receiveMail'], True)

    def test_nested_container_boolean_field_no_slug(self):
        profile = UserProfile.objects.create(user=self.user)
        setting = NotificationSetting.objects.create(user=profile, receiveMail=False)
        body = {
            'settings': {
                'http://happy-dev.fr/owl/#@id': setting.urlid,
                'receiveMail': True
            },
            "@context": {"@vocab": "http://happy-dev.fr/owl/#",
                         "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                         "rdfs": "http://www.w3.org/2000/01/rdf-schema#", "ldp": "http://www.w3.org/ns/ldp#",
                         "foaf": "http://xmlns.com/foaf/0.1/", "name": "rdfs:label",
                         "acl": "http://www.w3.org/ns/auth/acl#", "permissions": "acl:accessControl",
                         "mode": "acl:mode", "geo": "http://www.w3.org/2003/01/geo/wgs84_pos#", "lat": "geo:lat",
                         "lng": "geo:long"}
        }

        response = self.client.patch('/userprofiles/{}/'.format(profile.slug),
                                     data=json.dumps(body),
                                     content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['settings']['receiveMail'], True)

    # variation where the lookup_field for NotificationSetting (pk) is provided
    def test_nested_container_boolean_field_with_slug(self):
        profile = UserProfile.objects.create(user=self.user)
        setting = NotificationSetting.objects.create(user=profile, receiveMail=False)
        body = {
            'settings': {
                'pk': setting.pk,
                'http://happy-dev.fr/owl/#@id': setting.urlid,
                'receiveMail': True
            },
            "@context": {"@vocab": "http://happy-dev.fr/owl/#",
                         "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                         "rdfs": "http://www.w3.org/2000/01/rdf-schema#", "ldp": "http://www.w3.org/ns/ldp#",
                         "foaf": "http://xmlns.com/foaf/0.1/", "name": "rdfs:label",
                         "acl": "http://www.w3.org/ns/auth/acl#", "permissions": "acl:accessControl",
                         "mode": "acl:mode", "geo": "http://www.w3.org/2003/01/geo/wgs84_pos#", "lat": "geo:lat",
                         "lng": "geo:long"}
        }

        response = self.client.patch('/userprofiles/{}/'.format(profile.slug),
                                     data=json.dumps(body),
                                     content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['settings']['receiveMail'], True)

    def test_update_container_twice_nested_view(self):
        invoice = Invoice.objects.create(title='test')
        pre_existing_batch = Batch.objects.create(title='batch1', invoice=invoice)
        pre_existing_task = Task.objects.create(title='task1', batch=pre_existing_batch)

        base_url = settings.BASE_URL

        body = {
            "@id": "{}/invoices/{}/".format(base_url, invoice.pk),
            "http://happy-dev.fr/owl/#title": "new",
            "http://happy-dev.fr/owl/#batches": [
                {
                    "@id": "{}/batchs/{}/".format(base_url, pre_existing_batch.pk),
                    "http://happy-dev.fr/owl/#title": "new",
                    "http://happy-dev.fr/owl/#tasks": [
                        {
                            "@id": "{}/tasks/{}/".format(base_url, pre_existing_task.pk),
                            "http://happy-dev.fr/owl/#title": "new"
                        },
                        {
                            "http://happy-dev.fr/owl/#title": "tache 2"
                        }
                    ]
                },
                {
                    "http://happy-dev.fr/owl/#title": "z",
                }
            ]
        }

        response = self.client.put('/invoices/{}/'.format(invoice.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

        self.assertEquals(response.data['title'], "new")
        self.assertEquals(response.data['@id'], invoice.urlid)

        invoice = Invoice.objects.get(pk=invoice.pk)
        self.assertIs(invoice.batches.count(), 2)
        batches = invoice.batches.all().order_by('title')
        self.assertEquals(batches[0].title, "new")
        self.assertEquals(batches[0].urlid, pre_existing_batch.urlid)
        self.assertEquals(batches[1].title, "z")

        self.assertIs(batches[0].tasks.count(), 2)
        tasks = batches[0].tasks.all().order_by('title')
        self.assertEquals(tasks[0].title, "new")
        self.assertEquals(tasks[0].pk, pre_existing_task.pk)
        self.assertEquals(tasks[1].title, "tache 2")

    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/333
    '''def test_update_container_nested_view(self):
        circle = Circle.objects.create(name='test')
        pre_existing = CircleMember.objects.create(user=self.user, circle=circle, is_admin=False)
        another_user = get_user_model().objects.create_user(username='u2', email='u2@b.com', password='pw')

        body = {
            "@id": "{}/circles/{}/".format(settings.BASE_URL, circle.pk),
            "http://happy-dev.fr/owl/#name": "Updated Name",
            "http://happy-dev.fr/owl/#members": {
                "ldp:contains": [
                    {"@id": "{}/circle-members/{}/".format(settings.BASE_URL, pre_existing.pk),
                     "http://happy-dev.fr/owl/#is_admin": True},
                    {"http://happy-dev.fr/owl/#user": {"@id": another_user.urlid},
                     "http://happy-dev.fr/owl/#is_admin": False},
                ]
            }
        }

        response = \
            self.client.put('/circles/{}/'.format(circle.pk), data=json.dumps(body), content_type='application/ld+json')
        print(str(self.user.urlid))
        print(str(response.data))
        self.assertEqual(response.status_code, 200)

        self.assertEquals(response.data['name'], circle.name)
        self.assertEqual(response.data['@id'], circle.urlid)
        self.assertIs(CircleMember.objects.count(), 2)
        self.assertIs(circle.members.count(), 2)
        self.assertIs(circle.team.count(), 2)

        members = circle.members.all().order_by('pk')
        self.assertEqual(members[0].user, self.user)
        self.assertEqual(members[0].urlid, pre_existing.urlid)
        self.assertEqual(members[0].pk, pre_existing.pk)
        self.assertEqual(members[0].is_admin, True)
        self.assertEqual(members[1].user, another_user)
        self.assertEqual(members[1].is_admin, False)'''




from django.conf import settings
from django.db.models import Q
from rest_framework.filters import BaseFilterBackend
from rest_framework_guardian.filters import ObjectPermissionsFilter
from djangoldp.utils import is_anonymous_user


class LDPPermissionsFilterBackend(ObjectPermissionsFilter):
    """
    Default FilterBackend for LDPPermissions. If user does not have model-level permissions, filters by
    Django-Guardian's get_objects_for_user
    """

    def filter_queryset(self, request, queryset, view):
        from djangoldp.models import Model
        from djangoldp.permissions import LDPPermissions, ModelConfiguredPermissions

        # compares the requirement for GET, with what the user has on the container
        configured_permission_classes = getattr(view.model._meta, 'permission_classes', [LDPPermissions])
        for permission_class in [p() for p in configured_permission_classes]:
            # inherits from LDPBasePermissions
            if hasattr(permission_class, 'has_container_permission') and \
                permission_class.has_container_permission(request, view):
                return queryset

        # the user did not have permission on the container, so now we filter the queryset for permissions on the object
        if not is_anonymous_user(request.user):
            # those objects I have by grace of group or object
            # first figure out if the superuser has special permissions (important to the implementation in superclass)
            from djangoldp.models import Model
            anon_perms, auth_perms, owner_perms, superuser_perms = Model.get_permission_settings(view.model)
            #if no view permission for superuser, set the shortcut to False
            self.shortcut_kwargs['with_superuser'] = ('view' in superuser_perms)
            filtered_queryset = super().filter_queryset(request, queryset, view)

            # those objects I have by grace of being owner
            if 'view' in owner_perms:
                if getattr(view.model._meta, 'owner_field', None) is not None:
                    return (filtered_queryset | queryset.filter(**{view.model._meta.owner_field: request.user})).distinct()
                if getattr(view.model._meta, 'owner_urlid_field', None) is not None:
                    return (filtered_queryset | queryset.filter(**{view.model._meta.owner_urlid_field: request.user.urlid})).distinct()
            return filtered_queryset

        # user is anonymous without anonymous permissions
        return view.model.objects.none()


class LocalObjectFilterBackend(BaseFilterBackend):
    """
    Filter which removes external objects (federated backlinks) from the queryset
    For querysets which should only include local objects
    """
    def filter_queryset(self, request, queryset, view):
        return queryset.filter(urlid__startswith=settings.SITE_URL)


class LocalObjectOnContainerPathBackend(LocalObjectFilterBackend):
    """
    Override of LocalObjectFilterBackend which removes external objects when the view requested
    is the model container path
    """
    def filter_queryset(self, request, queryset, view):
        from djangoldp.models import Model

        if issubclass(view.model, Model) and request.path_info == view.model.get_container_path():
            return super(LocalObjectOnContainerPathBackend, self).filter_queryset(request, queryset, view)
        return queryset


class SearchByQueryParamFilterBackend(BaseFilterBackend):
    """
    Applies search fields in the request query params to the queryset
    """
    def filter_queryset(self, request, queryset, view):

        # the model fields on which to perform the search
        search_fields = request.GET.get('search-fields', None)
        # the terms to search the fields for
        search_terms = request.GET.get('search-terms', None)
        # the method of search to apply to the model fields
        search_method = request.GET.get('search-method', "basic")
        # union or intersection
        search_policy = request.GET.get('search-policy', 'union')

        # check if there is indeed a search requested
        if search_fields is None or search_terms is None:
            return queryset

        def _construct_search_query(search):
            '''Utility function pipes many Django Query objects'''
            search_query = []

            for idx, s in enumerate(search):
                if idx > 0:

                    # the user has indicated to make a union of all query results
                    if search_policy == "union":
                        search_query = search_query | Q(**s)

                    # the user has indicated to make an intersection of all query results
                    else:
                        search_query = search_query & Q(**s)

                    continue

                search_query = Q(**s)

            return search_query

        search_fields = search_fields.split(',')

        if search_method == "basic":
            search = []

            for s in search_fields:
                query = {}
                query["{}__contains".format(s)] = search_terms
                search.append(query)

            queryset = queryset.filter(_construct_search_query(search))

        elif search_method == "ibasic":
            # NOTE: to use, see https://stackoverflow.com/questions/54071944/fielderror-unsupported-lookup-unaccent-for-charfield-or-join-on-the-field-not
            unaccent_extension = getattr(settings, 'SEARCH_UNACCENT_EXTENSION', False) and 'django.contrib.postgres' in settings.INSTALLED_APPS
            middle_term = '__unaccent' if unaccent_extension else ''

            search = []

            for s in search_fields:
                query = {}
                query["{}{}__icontains".format(s, middle_term)] = search_terms
                search.append(query)

            queryset = queryset.filter(_construct_search_query(search))

        elif search_method == "exact":
            search = []

            for s in search_fields:
                query = {}
                query["{}__exact".format(s)] = search_terms
                search.append(query)

            queryset = queryset.filter(_construct_search_query(search))

        return queryset



import threading
import json
import time
import copy
import requests
from queue import Queue
from requests.exceptions import Timeout, ConnectionError
from urllib.parse import urlparse
from django.contrib.auth import get_user_model
from django.db.models.signals import post_save, post_delete, m2m_changed
from django.dispatch import receiver, Signal
from django.conf import settings
from rest_framework.utils import model_meta

from djangoldp.models import Model, Follower, ScheduledActivity
from djangoldp.models import Activity as ActivityModel

import logging


logger = logging.getLogger('djangoldp')
BACKLINKS_ACTOR = {
    "type": "Service",
    "name": "Backlinks Service"
}

SCHEDULER_SETTINGS = {
    'apscheduler.timezone': getattr(settings, 'TIME_ZONE', 'UTC'),
}

MAX_ACTIVITY_RESCHEDULES = getattr(settings, 'MAX_ACTIVITY_RESCHEDULES', 3)
DEFAULT_BACKOFF_FACTOR = getattr(settings, 'DEFAULT_BACKOFF_FACTOR', 1)
DEFAULT_ACTIVITY_DELAY = getattr(settings, 'DEFAULT_ACTIVITY_DELAY', 0.1)
DEFAULT_REQUEST_TIMEOUT = getattr(settings, 'DEFAULT_REQUEST_TIMEOUT', 10)
MAX_RECORDS_ACTIVITY_CACHE = getattr(settings, 'MAX_RECORDS_ACTIVITY_CACHE', 10000)


activity_sending_finished = Signal()


class ActivityInMemoryCache:
    '''
    {
        'urlid': {
            # if an object is sent without a urlid it is not cached
            'object_urlid': {
                # for create/update, just interested in the most recent activity on this external id
                './': ACTIVITY

                # for add/remove, we're interested also in the target (container_id)
                'circles': ACTIVITY
                'circle-members': ACTIVITY
            }
        }
    }
    '''

    def __init__(self):
        self.cache = {
        }

    def reset(self):
        self.cache = {
        }

    def has(self, urlid, object_id, target_id=None):
        if urlid not in self.cache or object_id not in self.cache[urlid]:
            return False
    
        if target_id is None:
            target_id = './'
        
        return target_id in self.cache[urlid][object_id]

    def get(self, urlid, object_id, target_id=None):
        if target_id is None:
            target_id = './'

        if self.has(urlid, object_id, target_id):
            return self.cache[urlid][object_id][target_id]
        else:
            return None

    def set(self, urlid, object_id, target_id=None, value=None):
        if MAX_RECORDS_ACTIVITY_CACHE == 0:
            return

        if len(self.cache.keys()) > MAX_RECORDS_ACTIVITY_CACHE:
            self.reset()

        if target_id is None:
            target_id = './'

        if urlid not in self.cache:
            self.cache[urlid] = {}
        
        if object_id not in self.cache[urlid]:
            self.cache[urlid][object_id] = {}
        
        self.cache[urlid][object_id][target_id] = copy.deepcopy(value)

    def invalidate(self, urlid, object_id=None, target_id=None):
        # can clear the cache for an entire record or at any level in the cache
        if object_id is not None:
            if target_id is not None:
                self.cache[urlid][object_id].pop(target_id, None)
            else:
                self.cache[urlid].pop(object_id, None)
        else:
            self.cache.pop(urlid, None)


# used to minimise the activity traffic to necessary activities,
# preferred over a database solution which is slower
ACTIVITY_CACHE = ActivityInMemoryCache()
ACTIVITY_SAVING_SETTING = getattr(settings, 'STORE_ACTIVITIES', 'ERROR')


class ActivityQueueService:
    '''Manages an asynchronous queue for Activity format messages'''
    initialized = False
    queue = None

    @classmethod
    def revive_activities(cls):
        '''re-schedules all ScheduledActivities to the queue'''
        with cls.queue.mutex:
            cls.queue.queue.clear()

        scheduled = ScheduledActivity.objects.all()
        for activity in scheduled:
            if activity.external_id is not None:
                cls.resend_activity(str(activity.external_id), activity, failed=False)
            else:
                activity.delete()

    @classmethod
    def start(cls):
        '''
        method checks if there are scheduled activities on the queue and starts them up
        Important: this method should only be called in start-up, when you know there are not queue tasks running
        otherwise duplicate activities may be sent
        '''
        def queue_worker(queue):
            while True:
                # wait for queue item to manifest
                item = queue.get()
                time.sleep(item[2])
                cls._activity_queue_worker(item[0], item[1])
                cls.queue.task_done()

        if not cls.initialized:
            cls.initialized = True

            # initialise the queue worker - infinite maxsize
            cls.queue = Queue(maxsize=0)
            t = threading.Thread(target=queue_worker, args=[cls.queue])
            t.setDaemon(True)
            t.start()

            cls.revive_activities()

    @classmethod
    def do_post(cls, url, activity, auth=None, timeout=DEFAULT_REQUEST_TIMEOUT):
        '''
        makes a POST request to url, passing activity
        :returns: response from server
        :raises: Timeout or ConnectionError if the post could not be made
        '''
        headers = {'Content-Type': 'application/ld+json'}
        logger.debug('[Sender] sending Activity... ' + str(activity))

        if getattr(settings, 'DISABLE_OUTBOX', False) == 'DEBUG':
            return {'data': {}}
        return requests.post(url, data=json.dumps(activity), headers=headers, timeout=timeout)

    @classmethod
    def _get_str_urlid(cls, obj):
        if obj is None:
            return None

        return obj if isinstance(obj, str) else obj['@id'] if '@id' in obj else None

    @classmethod
    def _save_activity_from_response(cls, response, url, scheduled_activity):
        '''
        wrapper to save a finished Activity based on the parameterised response
        :return: saved Activity object
        '''
        response_body = None

        if hasattr(response, 'text'):
            response_body = response.text
        response_location = getattr(response, "Location", None)
        status_code = getattr(response, "status_code", response['status_code'] if 'status_code' in response else None)
        success = str(status_code).startswith("2") if status_code is not None else False
        
        # strip some key info from the activity (standardise the datatype)
        activity_json = scheduled_activity.to_activitystream() if isinstance(scheduled_activity, ScheduledActivity) else scheduled_activity
        activity_type = scheduled_activity.type if hasattr(scheduled_activity, 'type') else scheduled_activity['type']

        # set the activity cache
        if 'object' in activity_json:
            object_id = cls._get_str_urlid(activity_json['object'])
            
            if object_id is not None:
                target_origin = cls._get_str_urlid(activity_json.get('target', activity_json.get('origin', None)))
                ACTIVITY_CACHE.set(url, object_id, target_origin, activity_json)

        # save the activity if instructed to do so
        if ACTIVITY_SAVING_SETTING == 'VERBOSE' or (not success and ACTIVITY_SAVING_SETTING == 'ERROR'):
            return cls._save_sent_activity(activity_json, ActivityModel, success=success,
                                           external_id=url, type=activity_type,
                                           response_location=response_location, response_code=str(status_code),
                                           response_body=response_body)

    @classmethod
    def _attempt_failed_reschedule(cls, url, scheduled_activity, backoff_factor):
        '''
        either re-schedules a failed activity or saves its failure state, depending on the number of fails and the
        fail policy (MAX_ACTIVITY_RESCHEDULES)
        :return: True if it was able to reschedule
        '''
        if scheduled_activity.failed_attempts < MAX_ACTIVITY_RESCHEDULES:
            backoff = backoff_factor * (2 ** (scheduled_activity.failed_attempts - 1))
            cls.resend_activity(url, scheduled_activity, backoff)
            return True

        # no retries left, save the failure state
        logger.error('Failed to deliver backlink to ' + str(url) + ' after retrying ' +
                     str(MAX_ACTIVITY_RESCHEDULES) + ' times')

        cls._save_sent_activity(scheduled_activity.to_activitystream(), ActivityModel, success=False, external_id=url,
                                type=scheduled_activity.type, response_code='408')
        return False

    @classmethod
    def _dispatch_activity_sending_finished(cls, response, saved_activity):
        '''sends a 'activity_sending_finished' signal to receivers'''
        activity_sending_finished.send(sender=cls, response=response, saved_activity=saved_activity)

    @classmethod
    def _send_activity(cls, url, scheduled_activity, auth=None, backoff_factor=DEFAULT_BACKOFF_FACTOR):
        '''
        makes a POST request to url, passing ScheduledActivity instance. reschedules if needed
        :param backoff_factor: a factor to use in the extension of waiting for retries. Used both in the RetryStrategy
        of requests.post and in rescheduling an activity which timed out
        '''
        response = None
        activity = scheduled_activity.to_activitystream()
        try:
            response = cls.do_post(url, activity, auth)
        except (Timeout, ConnectionError):
            if cls._attempt_failed_reschedule(url, scheduled_activity, backoff_factor):
                # successfully rescheduled, so skip cleanup for now
                return
        except Exception as e:
            logger.error('Failed to deliver backlink to ' + str(url) + ', was attempting ' + str(activity) +
                         str(e.__class__) + ': ' + str(e))

        saved = None
        if response is not None:
            saved = cls._save_activity_from_response(response, url, scheduled_activity)
        scheduled_activity.delete()

        # emit activity finished event
        cls._dispatch_activity_sending_finished(response, saved)

    @classmethod
    def _activity_queue_worker(cls, url, scheduled_activity):
        '''
        Worker for sending a scheduled activity on the queue. Decides whether to send the activity and then passes to
        _send_activity if it is worth it
        '''

        def get_related_activities(type):
            '''returns a list of activity types which should be considered a "match" with the parameterised type'''
            if type is None:
                return []
            type = type.lower()
            group_a = ['create', 'update', 'delete', 'creation', 'deletion']
            groub_b = ['add', 'remove']

            if type in group_a:
                return group_a
            if type in groub_b:
                return groub_b
            return []

        types = get_related_activities(scheduled_activity.type)
        if len(types) > 0:
            scheduled = ScheduledActivity.objects.filter(external_id=scheduled_activity.external_id,
                                                         created_at__gt=scheduled_activity.created_at,
                                                         type__in=types)

            # filter to scheduled activities on the same object
            scheduled = [s for s in scheduled if cls._is_same_object_target(s, scheduled_activity)]

            if len(scheduled) > 0:
                scheduled_activity.delete()
                return

        if scheduled_activity.type == 'update' and not cls._update_is_new(url, scheduled_activity):
            scheduled_activity.delete()
            return

        if scheduled_activity.type in ['add', 'remove'] and not cls._add_remove_is_new(url, scheduled_activity):
            scheduled_activity.delete()
            return

        cls._send_activity(url, scheduled_activity)

    @classmethod
    def _is_same_object_target(cls, activity_a, activity_b):
        def get_property(object, property, default=None):
            ret = object.get(property, default)
            if isinstance(ret, dict):
                ret = ret.get('@id', ret.get('name', None))
            return ret

        a = activity_a.to_activitystream()
        b = activity_b.to_activitystream()

        if get_property(a, 'object') != get_property(b, 'object'):
            return False

        return get_property(a, 'target', a.get('origin', None)) == get_property(b, 'target', b.get('origin', None))

    @classmethod
    def _update_is_new(cls, url, scheduled_activity):
        '''auxiliary function which validates if a scheduled update holds new information, compared to a past success'''
        def ordered(obj):
            '''recursively sorts nested dictionary objects to standardise ordering in comparison'''
            if isinstance(obj, dict):
                return sorted((k, ordered(v)) for k, v in obj.items())
            else:
                return obj

        def no_new_changes(old_activity, new_activity):
            '''returns False if the two activities are equivalent'''
            return ordered(old_activity['object']) == ordered(new_activity['object'])

        def get_most_recent_sent_activity(external_id, object_id):
            return ACTIVITY_CACHE.get(external_id, object_id)

        # str objects will have to be checked manually by the receiver
        new_activity = scheduled_activity.to_activitystream()
        if 'object' not in new_activity or isinstance(new_activity['object'], str) or \
             '@id' not in new_activity['object']:
        
            return True

        old_activity = get_most_recent_sent_activity(url, new_activity['object']['@id'])

        if old_activity is None:
            return True

        if no_new_changes(old_activity, new_activity):
            return False
        return True

    @classmethod
    def _add_remove_is_new(cls, url, scheduled_activity):
        '''auxiliary function validates if the receiver does not know about this Add/Remove activity'''
        def get_most_recent_sent_activity(source_obj, source_target_origin):
            obj_id = cls._get_str_urlid(source_obj)
            target_id = cls._get_str_urlid(source_target_origin)

            return ACTIVITY_CACHE.get(url, obj_id, target_id)

        new_activity = scheduled_activity.to_activitystream()
        new_obj = new_activity.get('object', None)
        new_target_origin = new_activity.get('target', new_activity.get('origin', None))

        # bounds checking
        if new_obj is None or new_target_origin is None or \
            (not isinstance(new_obj, str) and '@id' not in new_obj) or \
            (not isinstance(new_target_origin, str) and '@id' not in new_target_origin):
            return True

        #  if most recent is the same type of activity as me, it's not new
        old_activity = get_most_recent_sent_activity(new_obj, new_target_origin)

        if old_activity is not None and old_activity['type'].lower() == scheduled_activity.type.lower():
            return False
        return True

    @classmethod
    def _push_to_queue(cls, url, scheduled_activity, delay=DEFAULT_ACTIVITY_DELAY):
        '''wrapper to check for singleton initialization before pushing'''
        if not cls.initialized:
            cls.start()
        cls.queue.put([url, scheduled_activity, delay])

    @classmethod
    def resend_activity(cls, url, scheduled_activity, failed=True):
        '''
        a variation of send_activity for ScheduledActivity objects
        :param url: the recipient url inbox
        :param scheduled_activity: a ScheduledActivity object for sending
        :param failed: set to True to increment scheduled_activity.failed_attempts, to keep track of the number of resends
        '''
        if failed:
            scheduled_activity.failed_attempts = scheduled_activity.failed_attempts + 1
            scheduled_activity.save()

        cls._push_to_queue(url, scheduled_activity)

    @classmethod
    def send_activity(cls, url, activity, auth=None, delay=DEFAULT_ACTIVITY_DELAY):
        '''
        saves a ScheduledActivity for the parameterised activity and passes it to the queue
        :param url: the recipient url inbox
        :param activity: an Activity to send
        '''
        if getattr(settings, 'DISABLE_OUTBOX', False) is not False:
            if getattr(settings, 'DISABLE_OUTBOX') == 'DEBUG':
                cls._save_sent_activity(activity, ActivityModel, external_id=url, success=True, type=activity.get('type', None),
                                        response_code='201')
            return

        # schedule the activity
        scheduled = cls._save_sent_activity(activity, ScheduledActivity, external_id=url, type=activity.get('type', None))
        cls._push_to_queue(url, scheduled, delay)

    @classmethod
    def _save_sent_activity(cls, activity, model_represenation=ActivityModel, success=False, external_id=None, type=None,
                            response_location=None, response_code=None, local_id=None, response_body=None):
        '''
        Auxiliary function saves a record of parameterised activity
        :param model_represenation: the model class which should be used to store the activity. Defaults to djangol.Activity, must be a subclass
        '''
        payload = bytes(json.dumps(activity), "utf-8")
        if response_body is not None:
            response_body = bytes(json.dumps(response_body), "utf-8")
        if local_id is None:
            local_id = settings.SITE_URL + "/outbox/"
        if type is not None:
            type = type.lower()
        elif 'type' in activity and isinstance(activity.get('type'), str):
            type = activity.get('type').lower()
        obj = model_represenation.objects.create(local_id=local_id, payload=payload, success=success,
                                                 external_id=external_id, type=type, response_location=response_location,
                                                 response_code=response_code, response_body=response_body)
        return obj


class ActivityPubService(object):
    '''A service aiding the construction and sending of ActivityStreams notifications'''

    @classmethod
    def build_object_tree(cls, instance):
        '''builds a depth 1 object tree from a parameterised instance, with each branch being an object's urlid and RDF type'''
        model = type(instance)
        info = model_meta.get_field_info(model)

        if not hasattr(instance, 'urlid'):
            return

        obj = {
            "@type": Model.get_model_rdf_type(model),
            "@id": instance.urlid
        }
        if obj['@type'] is None:
            return

        # append relations
        for field_name, relation_info in info.relations.items():
            if not relation_info.to_many:
                value = getattr(instance, field_name, None)
                if value is None:
                    continue

                sub_object = {
                    "@id": value.urlid,
                    "@type": Model.get_model_rdf_type(type(value))
                }

                if sub_object['@type'] is None:
                    continue

                obj[field_name] = sub_object

        return obj

    @classmethod
    def get_actor_from_user_instance(cls, user):
        '''Auxiliary function returns valid Actor object from parameterised user instance, None if parameter invalid'''
        if isinstance(user, get_user_model()) and hasattr(user, 'urlid'):
            return {
                '@type': 'foaf:user',
                '@id': user.urlid
            }
        return None

    @classmethod
    def discover_inbox(cls, target_id):
        '''a method which discovers the inbox of the target resource'''
        url = urlparse(target_id)
        return target_id.replace(url.path, "/") + "inbox/"

    @classmethod
    def build_activity(self, actor, obj, activity_type='Activity', **kwargs):
        '''Auxiliary function returns an activity object with kwargs in the body'''
        res = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                settings.LDP_RDF_CONTEXT
            ],
            "type": activity_type,
            "actor": actor,
            "object": obj
        }

        for kwarg in kwargs:
            res.update({kwarg: kwargs[kwarg]})

        return res

    @classmethod
    def send_add_activity(cls, actor, obj, target):
        '''
        Sends an Add activity
        :param actor: a valid Actor object
        :param obj: a valid ActivityStreams Object
        :param target: an object representing the target collection
        '''
        summary = str(obj['@id']) + " was added to " + str(target['@id'])
        activity = cls.build_activity(actor, obj, activity_type='Add', summary=summary, target=target)

        # send request
        inbox = ActivityPubService.discover_inbox(target['@id'])
        ActivityQueueService.send_activity(inbox, activity)

    @classmethod
    def send_remove_activity(cls, actor, obj, origin):
        '''
        Sends a Remove activity
        :param actor: a valid Actor object, or a user instance
        :param obj: a valid ActivityStreams Object
        :param origin: the context the object has been removed from
        '''
        summary = str(obj['@id']) + " was removed from " + str(origin['@id'])
        activity = cls.build_activity(actor, obj, activity_type='Remove', summary=summary, origin=origin)

        # send request
        inbox = ActivityPubService.discover_inbox(origin['@id'])
        ActivityQueueService.send_activity(inbox, activity)

    @classmethod
    def send_create_activity(cls, actor, obj, inbox):
        '''
        Sends a Create activity
        :param actor: a valid Actor object, or a user instance
        :param obj: a valid ActivityStreams Object
        :param inbox: the inbox to send the activity to
        '''
        summary = str(obj['@id']) + " was created"
        activity = cls.build_activity(actor, obj, activity_type='Create', summary=summary)

        ActivityQueueService.send_activity(inbox, activity)

    @classmethod
    def send_update_activity(cls, actor, obj, inbox):
        '''
        Sends an Update activity
        :param actor: a valid Actor object, or a user instance
        :param obj: a valid ActivityStreams Object
        :param inbox: the inbox to send the activity to
        '''
        summary = str(obj['@id']) + " was updated"
        activity = cls.build_activity(actor, obj, activity_type='Update', summary=summary)

        ActivityQueueService.send_activity(inbox, activity)

    @classmethod
    def send_delete_activity(cls, actor, obj, inbox):
        '''
        Sends a Remove activity
        :param actor: a valid Actor object, or a user instance
        :param obj: a valid ActivityStreams Object
        :param inbox: the inbox to send the activity to
        '''
        summary = str(obj['@id']) + " was deleted"
        activity = cls.build_activity(actor, obj, activity_type='Delete', summary=summary)

        ActivityQueueService.send_activity(inbox, activity)

    @classmethod
    def get_related_externals(cls, sender, instance):
        '''Auxiliary function returns a set of urlids of distant resources connected to paramertised instance'''
        info = model_meta.get_field_info(sender)

        # bounds checking
        if not hasattr(instance, 'urlid') or Model.get_model_rdf_type(sender) is None:
            return set()

        # check each foreign key for a distant resource
        targets = set()
        for field_name, relation_info in info.relations.items():
            if not relation_info.to_many:
                value = getattr(instance, field_name, None)
                if value is not None and Model.is_external(value):
                    target_type = Model.get_model_rdf_type(type(value))

                    if target_type is None:
                        continue

                    targets.add(value.urlid)

        return targets

    @classmethod
    def get_target_inboxes(cls, urlids):
        '''Auxiliary function returns a set of inboxes, from a set of target object urlids'''
        inboxes = set()
        for urlid in urlids:
            inboxes.add(ActivityPubService.discover_inbox(urlid))
        return inboxes

    @classmethod
    def get_follower_inboxes(cls, object_urlid):
        '''Auxiliary function returns a set of inboxes, from the followers of parameterised object urlid'''
        inboxes = set(Follower.objects.filter(object=object_urlid).values_list('inbox', flat=True))
        return inboxes

    @classmethod
    def save_follower_for_target(cls, external_urlid, obj_id):
        inbox = ActivityPubService.discover_inbox(external_urlid)

        if not Follower.objects.filter(object=obj_id, follower=external_urlid).exists():
            Follower.objects.create(object=obj_id, inbox=inbox, follower=external_urlid,
                                    is_backlink=True)

    @classmethod
    def save_followers_for_targets(cls, external_urlids, obj_id):
        '''
        saves Follower objects for any external urlid which isn't already following the object in question
        :param external_urlids: list of external urlids to populate the follower inbox
        :param obj_id: object id to be followed
        '''
        existing_followers = Follower.objects.filter(object=obj_id).values_list('follower', flat=True)
        for urlid in external_urlids:
            if urlid not in existing_followers:
                Follower.objects.create(object=obj_id, inbox=ActivityPubService.discover_inbox(urlid),
                                        follower=urlid, is_backlink=True)

    @classmethod
    def remove_followers_for_resource(cls, external_urlid, obj_id):
        '''removes all followers which match the follower urlid, obj urlid combination'''
        inbox = ActivityPubService.discover_inbox(external_urlid)

        for follower in Follower.objects.filter(object=obj_id, follower=external_urlid,
                                                inbox=inbox, is_backlink=True):
            follower.delete()


@receiver([post_save])
def check_save_for_backlinks(sender, instance, created, **kwargs):
    if getattr(settings, 'SEND_BACKLINKS', True) and getattr(instance, 'allow_create_backlink', False) \
            and not Model.is_external(instance) \
            and getattr(instance, 'username', None) != 'hubl-workaround-493':
        external_urlids = ActivityPubService.get_related_externals(sender, instance)
        inboxes = ActivityPubService.get_follower_inboxes(instance.urlid)
        targets = set().union(ActivityPubService.get_target_inboxes(external_urlids), inboxes)

        if len(targets) > 0:
            obj = ActivityPubService.build_object_tree(instance)
            actor = BACKLINKS_ACTOR
            # Create Activity
            if created:
                for target in targets:
                    ActivityPubService.send_create_activity(actor, obj, target)
            # Update Activity
            else:
                for target in targets:
                    ActivityPubService.send_update_activity(actor, obj, target)

            # create Followers to update external resources of changes in future
            ActivityPubService.save_followers_for_targets(external_urlids, obj['@id'])


@receiver([post_delete])
def check_delete_for_backlinks(sender, instance, **kwargs):
    if getattr(settings, 'SEND_BACKLINKS', True) and getattr(instance, 'allow_create_backlink', False) \
            and getattr(instance, 'username', None) != 'hubl-workaround-493':
        targets = ActivityPubService.get_follower_inboxes(instance.urlid)

        if len(targets) > 0:
            for target in targets:
                ActivityPubService.send_delete_activity(BACKLINKS_ACTOR, {
                    "@id": instance.urlid,
                    "@type": Model.get_model_rdf_type(sender)
                }, target)

    # remove any Followers on this resource
    urlid = getattr(instance, 'urlid', None)
    if urlid is not None:
        for follower in Follower.objects.filter(object=urlid):
            follower.delete()


@receiver([m2m_changed])
def check_m2m_for_backlinks(sender, instance, action, *args, **kwargs):
    def build_targets(query_set):
        '''analyses parameterised queryset (removed or added members) for backlinks'''
        targets = []

        for obj in query_set:
            condition = Model.is_external(obj) and getattr(obj, 'allow_create_backlink', False)
            if action == "post_add":
                condition = condition and not getattr(obj, 'is_backlink', True)

            if condition:
                targets.append({
                    "@type": member_rdf_type,
                    "@id": obj.urlid
                })

        return targets

    if getattr(settings, 'SEND_BACKLINKS', True):
        member_model = kwargs['model']
        pk_set = kwargs['pk_set']

        # we can only send backlinks on pre_clear because on post_clear the objects are gone
        if action != "pre_clear" and pk_set is None:
            return
        member_rdf_type = Model.get_model_rdf_type(member_model)
        container_rdf_type = Model.get_model_rdf_type(type(instance))

        if member_rdf_type is None:
            return
        if container_rdf_type is None:
            return

        # build list of targets (models affected by the change)
        if action == "pre_clear":
            sender_info = model_meta.get_field_info(sender)
            for field_name, relation_info in sender_info.relations.items():
                if relation_info.related_model == member_model:
                    pk_set = sender.objects.all().values_list(field_name, flat=True)
        if pk_set is None or len(pk_set) == 0:
            return
        query_set = member_model.objects.filter(pk__in=pk_set)
        targets = build_targets(query_set)

        if len(targets) > 0:
            obj = {
                "@type": container_rdf_type,
                "@id": instance.urlid
            }
            if action == 'post_add':
                for target in targets:
                    ActivityPubService.send_add_activity(BACKLINKS_ACTOR, obj, target)
                    ActivityPubService.save_follower_for_target(target['@id'], obj['@id'])

            elif action == "post_remove" or action == "pre_clear":
                for target in targets:
                    ActivityPubService.send_remove_activity(BACKLINKS_ACTOR, obj, target)
                    ActivityPubService.remove_followers_for_resource(target['@id'], obj['@id'])



from copy import deepcopy
import random
from faker import Faker

'''
Contains definitions used in common by multiple scripts within this directory
'''
def generate_user(i, user_template):
    myFactory = Faker()
    user = deepcopy(user_template)
    email = myFactory.unique.email().split('@')
    email.insert(1, str(random.randint(0, 5)))
    email.insert(2, "@")
    email_str = "".join(email)
    user['pk'] = i
    user['fields']['username'] = myFactory.unique.user_name() + str(random.randint(0, 100))
    user['fields']['email'] = email_str
    user['fields']['first_name'] = myFactory.first_name()
    user['fields']['last_name'] = myFactory.last_name()
    return user


def generate_users(count, user_template, fixture=None, offset=0):
    if fixture is None:
        fixture = list()

    for i in range(count):
        j = offset + i
        user = generate_user(j, user_template)
        fixture.append(user)

    return fixture


def generate_project_member_and_user_pks(project_pk, offset, total_users, max_members_per_project):
    '''
    returns a generator of tuples (new project member PKs and selected user PKs)
    raises error if there are not enough users
    '''
    # we want to select a handful of random users
    # to save time we just select a random user within a safe range and then grab a bunch of adjacent users
    start_user_pk = random.randint(max(offset, 1), offset + (total_users - (max_members_per_project + 1)))
    if start_user_pk < offset:
        raise IndexError('not enough users!')
    
    for i in range(random.randint(1, max_members_per_project)):
        j = offset + (i + (project_pk * max_members_per_project))  # generate a unique integer id
        user_pk = start_user_pk + i  # select the next user

        yield (j, user_pk)


def generate_project_members(project_pk, fixture, offset, total_users):
    max_members_per_project = 10

    def generate_project_member(i, user_pk):
        return {
            'model': 'djangoldp_project.member',
            'pk': i,
            'fields': {
                'project': project_pk,
                'user': user_pk
            }
        }

    for (j, user_pk) in generate_project_member_and_user_pks(project_pk, offset, total_users, max_members_per_project):
        fixture.append(generate_project_member(j, user_pk))

    return fixture

def generate_skill(i, skill_template):
    myFactory = Faker()
    skill = deepcopy(skill_template)

    skill['pk'] = i
    skill['fields']['name'] = myFactory.unique.job()
    return skill


def generate_user_pks(skill_pk, offset, total_users, max_users_per_skill):
    '''
    returns a generator of tuples ()
    raises error if there are not enough users
    '''
    # we want to select a handful of random users
    # to save time we just select a random user within a safe range and then grab a bunch of adjacent users
    start_user_pk = random.randint(0, offset + (total_users - (max_users_per_skill + 1)))
    # if start_user_pk < offset:
    #     raise IndexError('not enough users!')
    
    for i in range(random.randint(1, max_users_per_skill)):
        j = offset + (i + (skill_pk * max_users_per_skill))  # generate a unique integer id
        user_pk = start_user_pk + i  # select the next user

        yield (j, user_pk)

def append_users_to_skill(skill, offset, total_users):
    max_users_per_skill = 250

    skill['fields']['users'] = []
    for (j, user_pk) in generate_user_pks(skill['pk'], offset, total_users, max_users_per_skill):
        skill['fields']['users'].append(user_pk)

    return skill

def generate_skills(count, skill_template, fixture=None, offset=0):
    if fixture is None:
        fixture = list()

    for i in range(count):
        j = offset + i
        skill = generate_skill(j, skill_template)

        skill = append_users_to_skill(skill, offset, count)
        fixture.append(skill)

    return fixture

def generate_project(i, project_template):
    project = deepcopy(project_template)
    project['pk'] = i

    return project


def append_members_to_project(project, offset, total_users):
    max_members_per_project = 10

    project['members'] = []
    for (j, user_pk) in generate_project_member_and_user_pks(project['pk'], offset, total_users, max_members_per_project):
        project['members'].append(user_pk)

    return project


def generate_projects(count, project_template, fixture=None, offset=0, production=True):
    if fixture is None:
        fixture = list()

    for i in range(count):
        j = offset + i
        project = generate_project(j, project_template)

        # project members using direct ManyToMany field. Generate them as a field on project
        if not production:
            project = append_members_to_project(project, offset, count)

        fixture.append(project)

        # project members using Member through model, generate them as separate in the fixture
        if production:
            # append random number of project members, max 10 for a single project
            generate_project_members(j, fixture, offset, count)

    return fixture



from django.db.models import options

__version__ = '3.0.8'

options.DEFAULT_NAMES += (
    'lookup_field', 'rdf_type', 'rdf_context', 'auto_author', 'auto_author_field', 'owner_field', 'owner_urlid_field',
    'view_set', 'container_path', 'permission_classes', 'serializer_fields', 'serializer_fields_exclude', 'empty_containers',
    'nested_fields', 'nested_fields_exclude', 'depth', 'anonymous_perms', 'authenticated_perms', 'owner_perms',
    'superuser_perms')



from django.conf import settings
from django.contrib.auth import _get_backends
from rest_framework.permissions import DjangoObjectPermissions
from djangoldp.utils import is_anonymous_user
from djangoldp.filters import LDPPermissionsFilterBackend


DEFAULT_DJANGOLDP_PERMISSIONS = ['add', 'change', 'delete', 'view', 'control']


class LDPBasePermission(DjangoObjectPermissions):
    """
    A base class from which all permission classes should inherit.
    Extends the DRF permissions class to include the concept of model-permissions, separate from the view, and to
    change to a system of outputting permissions sets for the serialization of WebACLs
    """
    # filter backends associated with the permissions class. This will be used to filter queryset in the (auto-generated)
    # view for a model, and in the serializing nested fields
    filter_backends = []
    # perms_map defines the permissions required for different methods
    perms_map = {
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': [],
        'HEAD': ['%(app_label)s.view_%(model_name)s'],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }

    def get_container_permissions(self, request, view, obj=None):
        """
        outputs a set of permissions of a given container. Used in the generation of WebACLs in LDPSerializer
        """
        return set()

    def get_object_permissions(self, request, view, obj):
        """
        outputs the permissions of a given object instance. Used in the generation of WebACLs in LDPSerializer
        """
        return set()

    def get_user_permissions(self, request, view, obj):
        '''
        returns a set of all model permissions and object permissions for given parameters
        You shouldn't override this function
        '''
        perms = self.get_container_permissions(request, view, obj)
        if obj is not None:
            return perms.union(self.get_object_permissions(request, view, obj))
        return perms

    def has_permission(self, request, view):
        """concerned with the permissions to access the _view_"""
        return True

    def has_container_permission(self, request, view):
        """
        concerned with the permissions to access the _model_
        in most situations you won't need to override this. It is primarily called by has_object_permission
        checked when POSTing to LDPViewSet
        """
        required_perms = self.get_required_permissions(request.method, view.model)
        return self.compare_permissions(required_perms, self.get_container_permissions(request, view))

    def has_object_permission(self, request, view, obj):
        """concerned with the permissions to access the _object_"""
        required_perms = self.get_required_permissions(request.method, view.model)
        return self.compare_permissions(required_perms, self.get_user_permissions(request, view, obj))

    def compare_permissions(self, required_perms, user_perms):
        '''returns True if all user_perms are in required_perms'''
        for perm in required_perms:
            if not perm.split('.')[-1].split('_')[0] in user_perms:
                return False
        return True

def select_container_permissions(request, obj, model, anonymous_perms, authenticated_perms, owner_perms, superuser_perms):
    from djangoldp.models import Model
    
    if is_anonymous_user(request.user):
        return set(anonymous_perms)
    else:
        if obj is not None and Model.is_owner(model, request.user, obj):
            perms = set(owner_perms)
        else:
            perms = set(authenticated_perms)
        if request.user.is_superuser:
            perms = perms.union(set(superuser_perms))
    return perms

class ModelConfiguredPermissions(LDPBasePermission):
    # *DEFAULT* model-level permissions for anon, auth and owner statuses
    anonymous_perms = ['view']
    authenticated_perms = ['inherit']
    owner_perms = ['inherit']
    # superuser has all permissions by default
    superuser_perms = getattr(settings, 'DEFAULT_SUPERUSER_PERMS', DEFAULT_DJANGOLDP_PERMISSIONS)

    def get_container_permissions(self, request, view, obj=None):
        '''analyses the Model's set anonymous, authenticated and owner_permissions and returns these'''
        perms = super().get_container_permissions(request, view, obj=obj)
        from djangoldp.models import Model
        if isinstance(view.model, Model):
            anonymous_perms, authenticated_perms, owner_perms, superuser_perms = view.model.get_permission_settings()
        else:
            anonymous_perms, authenticated_perms, owner_perms, superuser_perms = Model.get_permission_settings(view.model)
        return select_container_permissions(request, obj, view.model, anonymous_perms, authenticated_perms, owner_perms, superuser_perms)

    def has_permission(self, request, view):
        """concerned with the permissions to access the _view_"""
        if is_anonymous_user(request.user):
            if not self.has_container_permission(request, view):
                return False
        return True


class LDPObjectLevelPermissions(LDPBasePermission):
    def get_all_user_object_permissions(self, user, obj):
        return user.get_all_permissions(obj)


    def get_object_permissions(self, request, view, obj):
        '''overridden to append permissions from all backends given to the user (e.g. Groups and object-level perms)'''
        from djangoldp.models import Model

        model_name = Model.get_meta(view.model, 'model_name')

        perms = super().get_object_permissions(request, view, obj)

        if obj is not None and not is_anonymous_user(request.user):
            forbidden_string = "_" + model_name
            return perms.union(set([p.replace(forbidden_string, '') for p in
                                    self.get_all_user_object_permissions(request.user, obj)]))

        return perms


class SuperUserPermission(LDPBasePermission):
    filter_backends = []

    def get_container_permissions(self, request, view, obj=None):
        if request.user.is_superuser:
            return set(DEFAULT_DJANGOLDP_PERMISSIONS)
        return super().get_container_permissions(request, view, obj)

    def get_object_permissions(self, request, view, obj):
        if request.user.is_superuser:
            return set(DEFAULT_DJANGOLDP_PERMISSIONS)
        return super().get_object_permissions(request, view, obj)

    def has_permission(self, request, view):
        if request.user.is_superuser:
            return True
        return super().has_permission(request, view)

    def has_container_permission(self, request, view):
        if request.user.is_superuser:
            return True
        return super().has_container_permission(request, view)

    def has_object_permission(self, request, view, obj):
        if request.user.is_superuser:
            return True
        return super().has_object_permission(request, view, obj)


class LDPPermissions(LDPObjectLevelPermissions, ModelConfiguredPermissions):
    filter_backends = [LDPPermissionsFilterBackend]

    def get_all_user_object_permissions(self, user, obj):
        # if the super_user perms are no different from authenticated_perms, then we want to skip Django's auth backend
        restore_super = False
        if user.is_superuser:
            user.is_superuser = False
            restore_super = True

        perms = super().get_all_user_object_permissions(user, obj)

        user.is_superuser = restore_super
        return perms



from datetime import datetime
from django.contrib.auth import get_user_model
from rest_framework.test import APIRequestFactory, APIClient, APITestCase

from djangoldp.tests.models import Post, Invoice, JobOffer, Skill, Batch, DateModel, Circle, CircleMember, UserProfile
from djangoldp.serializers import GLOBAL_SERIALIZER_CACHE

class TestGET(APITestCase):

    def setUp(self):
        self.factory = APIRequestFactory()
        self.client = APIClient()
        self.ordered_fields = ['@context', '@type', '@id']
        setattr(Circle._meta, 'depth', 0)
        setattr(Circle._meta, 'empty_containers', [])

    def tearDown(self):
        GLOBAL_SERIALIZER_CACHE.reset()

    def test_get_resource(self):
        post = Post.objects.create(content="content")
        response = self.client.get('/posts/{}/'.format(post.pk), content_type='application/ld+json', HTTP_ORIGIN='http://localhost:8080/test/')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(response.data['content'], "content")
        self.assertIn('author', response.data)
        self.assertIn('@type', response.data)
        self.assertIn('permissions', response.data)


        # test headers returned
        self.assertEqual(response['Content-Type'], 'application/ld+json') 
        self.assertEqual(response['Accept-Post'], 'application/ld+json')
        self.assertEqual(response['Allow'], 'GET, PUT, PATCH, DELETE, HEAD, OPTIONS')
        self.assertEqual(response['Access-Control-Allow-Origin'], 'http://localhost:8080/test/')
        self.assertIn('DPoP', response['Access-Control-Allow-Headers'])

    def test_get_resource_urlid(self):
        user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                    password='glass onion')
        UserProfile.objects.create(user=user)
        post = Post.objects.create(content="content", author=user.userprofile)
        response = self.client.get('/posts/{}/'.format(post.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(response.data['content'], "content")
        self.assertEqual(response.data['author']['@id'], user.userprofile.urlid)

    def test_get_container(self):
        Post.objects.create(content="content")
        # federated object - should not be returned in the container view
        Post.objects.create(content="federated", urlid="https://external.com/posts/1/")
        response = self.client.get('/posts/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(1, len(response.data['ldp:contains']))
        self.assertIn('@type', response.data)
        self.assertIn('@type', response.data['ldp:contains'][0])
        self.assertNotIn('permissions', response.data['ldp:contains'][0])
        self.assertEquals(4, len(response.data['permissions'])) # configured anonymous permissions to give all

        Invoice.objects.create(title="content")
        response = self.client.get('/invoices/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('permissions', response.data)
        self.assertEquals(1, len(response.data['permissions']))  # read only

    def test_get_empty_container(self):
        response = self.client.get('/posts/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(0, len(response.data['ldp:contains']))

    def test_get_filtered_fields(self):
        skill = Skill.objects.create(title="Java", obligatoire="ok", slug="1")
        skill2 = Skill.objects.create(title="Java", obligatoire="ok", slug="2")
        skill3 = Skill.objects.create(urlid="http://happy-dev.hubl.fr/skills/1")
        job = JobOffer.objects.create(title="job", slug="1")
        job.skills.add(skill)
        job.skills.add(skill2)
        job.skills.add(skill3)
        job.save()
        response = self.client.get('/job-offers/{}/'.format(job.slug), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('recent_skills', response.data)
        self.assertEqual(response.data['recent_skills']['@id'], "http://happy-dev.fr/job-offers/1/recent_skills/")
        # the external resource should be serialized with its @id and @type.. and only these fields
        self.assertEqual(response.data['skills']['ldp:contains'][2]['@id'], "http://happy-dev.hubl.fr/skills/1")
        self.assertIn('@type', response.data['skills']['ldp:contains'][1])
        self.assertIn('@type', response.data['skills']['ldp:contains'][2])
        self.assertEqual(len(response.data['skills']['ldp:contains'][2].items()), 2)

    def test_get_reverse_filtered_fields(self):
        skill = Skill.objects.create(title="Java", obligatoire="ok", slug="1")
        job = JobOffer.objects.create(title="job", slug="1")
        job.skills.add(skill)
        job.save()
        response = self.client.get('/skills/{}/'.format(skill.slug), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('recent_jobs', response.data)
        self.assertEqual(response.data['recent_jobs']['@id'], "http://happy-dev.fr/skills/1/recent_jobs/")

    def test_get_virtual_field(self):
        skill = Skill.objects.create(title="Java", obligatoire="ok", slug="1")
        skill2 = Skill.objects.create(title="Java", obligatoire="ok", slug="2")
        job = JobOffer.objects.create(title="job", slug="1")
        job.skills.add(skill)
        job.skills.add(skill2)
        job.save()
        response = self.client.get('/job-offers/{}/'.format(job.slug), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('some_skill', response.data)
        self.assertEqual(response.data['some_skill']['@id'], skill.urlid)

    def test_get_nested(self):
        invoice = Invoice.objects.create(title="invoice")
        batch = Batch.objects.create(invoice=invoice, title="batch")
        distant_batch = Batch.objects.create(invoice=invoice, title="distant", urlid="https://external.com/batch/1/")
        response = self.client.get('/invoices/{}/batches/'.format(invoice.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(response.data['@id'], 'http://happy-dev.fr/invoices/{}/batches/'.format(invoice.pk))
        self.assertIn('permissions', response.data)
        self.assertEquals(len(response.data['ldp:contains']), 2)
        self.assertIn('@type', response.data['ldp:contains'][0])
        self.assertIn('@type', response.data['ldp:contains'][1])
        self.assertNotIn('permissions', response.data['ldp:contains'][0])
        self.assertNotIn('permissions', response.data['ldp:contains'][1])
        self.assertEquals(response.data['ldp:contains'][0]['invoice']['@id'], invoice.urlid)
        self.assertEqual(response.data['ldp:contains'][1]['@id'], distant_batch.urlid)

    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/335
    #  test getting a route with multiple nested fields (/job-offers/X/skills/Y/)
    '''def test_get_twice_nested(self):
        job = JobOffer.objects.create(title="job", slug="slug1")
        skill = Skill.objects.create(title='old', obligatoire='old', slug='skill1')
        job.skills.add(skill)
        self.assertEqual(job.skills.count(), 1)
        
        response = self.client.get('/job-offers/{}/skills/{}/'.format(job.slug, skill.slug))
        self.assertEqual(response.status_code, 200)'''

    def test_serializer_excludes(self):
        date = DateModel.objects.create(excluded='test', value=datetime.now())
        response = self.client.get('/dates/{}/'.format(date.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertNotIn('excluded', response.data.keys())

    def test_serializer_excludes_serializer_fields_set_also(self):
        setattr(DateModel._meta, 'serializer_fields', ['value', 'excluded'])
        date = DateModel.objects.create(excluded='test', value=datetime.now())
        response = self.client.get('/dates/{}/'.format(date.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertNotIn('excluded', response.data.keys())

    def _set_up_circle_and_user(self):
        circle = Circle.objects.create(name='test', description='test')
        user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                    password='glass onion')
        self.client.force_authenticate(user)
        CircleMember.objects.create(user=user, circle=circle)

    # tests for functionality allowing me to set containers to be serialized without content\
    # test for normal functioning (without setting)
    def test_empty_container_serialization_nested_serializer_no_empty(self):
        setattr(Circle._meta, 'depth', 1)
        self._set_up_circle_and_user()

        response = self.client.get('/circles/', content_type='application/ld+json')
        self.assertEqual(response.data['@type'], 'ldp:Container')
        self.assertIn('@id', response.data)
        self.assertIn('permissions', response.data)
        self.assertIn('members', response.data['ldp:contains'][0])
        self.assertEqual(response.data['ldp:contains'][0]['members']['@type'], 'ldp:Container')
        self.assertIn('@id', response.data['ldp:contains'][0]['members'])
        self.assertIn('ldp:contains', response.data['ldp:contains'][0]['members'])
        self.assertIn('permissions', response.data['ldp:contains'][0]['members'])

    # test for functioning with setting
    def test_empty_container_serialization_nested_serializer_empty(self):
        setattr(Circle._meta, 'depth', 1)
        setattr(Circle._meta, 'empty_containers', ['members'])
        self._set_up_circle_and_user()

        response = self.client.get('/circles/', content_type='application/ld+json')
        self.assertEqual(response.data['@type'], 'ldp:Container')
        self.assertIn('members', response.data['ldp:contains'][0])
        self.assertIn('@id', response.data['ldp:contains'][0]['members'])
        self.assertNotIn('@type', response.data['ldp:contains'][0]['members'])
        self.assertNotIn('permissions', response.data['ldp:contains'][0]['members'])
        self.assertNotIn('ldp:contains', response.data['ldp:contains'][0]['members'])

    # should serialize as normal on the nested viewset (directly asking for the container)
    # test for normal functioning (without setting)
    def test_empty_container_serialization_nested_viewset_no_empty(self):
        self._set_up_circle_and_user()

        response = self.client.get('/circles/1/members/', content_type='application/ld+json')
        self.assertEqual(response.data['@type'], 'ldp:Container')
        self.assertIn('@id', response.data)
        self.assertIn('ldp:contains', response.data)
        self.assertIn('permissions', response.data)
        self.assertIn('circle', response.data['ldp:contains'][0])

    # test for functioning with setting
    def test_empty_container_serialization_nested_viewset_empty(self):
        setattr(Circle._meta, 'empty_containers', ['members'])
        self._set_up_circle_and_user()

        response = self.client.get('/circles/1/members/', content_type='application/ld+json')
        self.assertEqual(response.data['@type'], 'ldp:Container')
        self.assertIn('@id', response.data)
        self.assertIn('ldp:contains', response.data)
        self.assertIn('permissions', response.data)
        self.assertIn('circle', response.data['ldp:contains'][0])

    # test for checking fields ordering
    def test_ordered_field(self):
        self._set_up_circle_and_user()
        response = self.client.get('/users/', content_type='application/ld+json')
        fields_to_test = [
            response.data.keys(),
            response.data['ldp:contains'][-1],
            response.data['ldp:contains'][-1]['circle_set']
        ]

        for test_fields in fields_to_test:
            test_fields = list(test_fields)
            o_f = [field for field in self.ordered_fields if field in test_fields]
            self.assertEquals(o_f, test_fields[:len(o_f)])
